// FunctionLibrary
//------------------------------------------------------------------------------

// Includes
//------------------------------------------------------------------------------
#include "FunctionLibrary.h"
#include <Tools/FBuild/FBuildCore/FBuild.h>
#include <Tools/FBuild/FBuildCore/Flog.h>
#include <Tools/FBuild/FBuildCore/BFF/BFFIterator.h>
#include <Tools/FBuild/FBuildCore/BFF/BFFParser.h>
#include <Tools/FBuild/FBuildCore/BFF/BFFStackFrame.h>
#include <Tools/FBuild/FBuildCore/BFF/BFFVariable.h>
#include <Tools/FBuild/FBuildCore/Graph/NodeGraph.h>
#include <Tools/FBuild/FBuildCore/Graph/DirectoryListNode.h>
#include <Tools/FBuild/FBuildCore/Graph/LibraryNode.h>
#include <Tools/FBuild/FBuildCore/Graph/ObjectNode.h>
#include <Tools/FBuild/FBuildCore/Graph/UnityNode.h>

// CONSTRUCTOR
//------------------------------------------------------------------------------
FunctionLibrary::FunctionLibrary()
: Function( "Library" )
{
}

// AcceptsHeader
//------------------------------------------------------------------------------
/*virtual*/ bool FunctionLibrary::AcceptsHeader() const
{
	return true;
}

// Commit
//------------------------------------------------------------------------------
/*virtual*/ bool FunctionLibrary::Commit( const BFFIterator & funcStartIter ) const
{
	// make sure all required variables are defined
	const BFFVariable * outputLib;
	const BFFVariable * compiler;
	const BFFVariable * compilerOptions;
	const BFFVariable * compilerOutputPath;
	const BFFVariable * librarian;
	const BFFVariable * librarianOptions;
	if ( !GetString( funcStartIter, outputLib, ".LibrarianOutput", true ) ||
		 !GetString( funcStartIter, compiler, ".Compiler", true ) ||
		 !GetString( funcStartIter, compilerOptions, ".CompilerOptions", true ) ||
		 !GetString( funcStartIter, compilerOutputPath, ".CompilerOutputPath", true ) ||
		 !GetString( funcStartIter, librarian, ".Librarian", true ) ||
		 !GetString( funcStartIter, librarianOptions, ".LibrarianOptions", true ) )
	{
		return false;
	}

	NodeGraph & ng = FBuild::Get().GetDependencyGraph();

	// find or create the compiler node
	Node * cn = ng.FindNode( compiler->GetValue() );
	FileNode * compilerNode = nullptr;
	if ( cn != nullptr )
	{
		if ( cn->IsAFile() == false )
		{
			Error::Error_1103_NotAFile( funcStartIter, this, "Compiler", cn->GetName(), cn->GetType() );
			return false;
		}
		compilerNode = cn->CastTo< FileNode >();
	}
	else
	{
		compilerNode = ng.CreateFileNode( compiler->GetValue() );
	}

	// Precompiled Header support
	ObjectNode * precompiledHeaderNode = nullptr;
	const BFFVariable * pchInputFile = nullptr;
	const BFFVariable * pchOutputFile = nullptr;
	const BFFVariable * pchOptions = nullptr;
	if ( !GetString( funcStartIter, pchInputFile, ".PCHInputFile" ) ||
		 !GetString( funcStartIter, pchOutputFile, ".PCHOutputFile" ) ||
		 !GetString( funcStartIter, pchOptions, ".PCHOptions" ) )
	{
		return false;
	}
	if ( pchInputFile ) 
	{
		if ( !pchOutputFile || !pchOptions )
		{
			Error::Error_1300_MissingPCHArgs( funcStartIter, this );
			return false;
		}
		Node * pchInputNode = ng.FindNode( pchInputFile->GetValue() );
		if ( pchInputNode )
		{
			// is it a file?
			if ( pchInputNode->IsAFile() == false )
			{
				Error::Error_1103_NotAFile( funcStartIter, this, "PCHInputFile", pchInputNode->GetName(), pchInputNode->GetType() );
				return false;
			}
		}
		else
		{
			// Create input node
			pchInputNode = ng.CreateFileNode( pchInputFile->GetValue() );
		}

		if ( ng.FindNode( pchOutputFile->GetValue() ) )
		{
			Error::Error_1301_AlreadyDefinedPCH( funcStartIter, this, pchOutputFile->GetValue().Get() );
			return false;
		}

		uint32_t flags( 0 );
		if ( compilerNode->GetName().EndsWith( "\\cl.exe" ) )
		{
			flags |= ObjectNode::FLAG_MSVC;
		}
		else if ( compilerNode->GetName().EndsWith( "gcc" ) )
		{
			flags |= ObjectNode::FLAG_GCC;
		}

		precompiledHeaderNode = ng.CreateObjectNode( pchOutputFile->GetValue(),
													 pchInputNode,
													 compilerNode,
													 pchOptions->GetValue(),
													 nullptr,
													 flags );
	}

	Array< Node * > staticDeps( 32, true );

	// do we want to build a files in a directory?
	const BFFVariable * inputPath = BFFStackFrame::GetVar( ".CompilerInputPath" );
	if ( inputPath )
	{
		// get the optional pattern and recurse options related to InputPath
		AStackString<> defaultWildCard( "*.cpp" );
		const BFFVariable * patternVar = BFFStackFrame::GetVar( ".CompilerInputPattern" );
		const AString & pattern = patternVar ? patternVar->GetValue() : defaultWildCard;

		// recursive?  default to true
		const BFFVariable * recurseVar = BFFStackFrame::GetVar( ".CompilerInputPathRecurse" );
		bool recurse = ( ( recurseVar == nullptr ) || ( recurseVar->GetValue() == "true" ) );

		// Support an exclusion path
		const BFFVariable * excludePathVar = BFFStackFrame::GetVar( ".CompilerInputExcludePath" );
		const AString & excludePath = excludePathVar ? excludePathVar->GetValue() : AString::GetEmpty();

		// clean paths
		AStackString< 512 > fullPath;
		NodeGraph::CleanPath( inputPath->GetValue(), fullPath );
		if ( fullPath.EndsWith( '\\' ) == false )
		{
			fullPath += '\\';
		}

		AStackString< 512 > fullExcludePath;
		if ( excludePath.IsEmpty() == false )
		{
			NodeGraph::CleanPath( excludePath, fullExcludePath );
			if ( fullExcludePath.EndsWith( '\\' ) == false )
			{
				fullExcludePath += '\\';
			}
		}

		// get node for the dir we depend on
		AStackString<> name;
		DirectoryListNode::FormatName( fullPath, pattern, recurse, fullExcludePath, name );
		Node * staticDep = ng.FindNode( name );
		if ( staticDep == nullptr )
		{
			staticDep = ng.CreateDirectoryListNode( name,
													fullPath,
													pattern,
													recurse,
													fullExcludePath );
		}
		else if ( staticDep->GetType() != Node::DIRECTORY_LIST_NODE )
		{
			Error::Error_1102_UnexpectedType( funcStartIter, this, "CompilerInputPath", staticDep->GetName(), staticDep->GetType(), Node::DIRECTORY_LIST_NODE );
			return false;
		}

		staticDeps.Append( staticDep );
	}

	// do we want to build a specific list of files?
	const BFFVariable * inputFiles = BFFStackFrame::GetVar( ".CompilerInputFiles" );
	if ( inputFiles )
	{
		if ( inputFiles->IsString() )
		{
			// get node for the file we depend on
			Node * staticDep = ng.FindNode( inputFiles->GetValue() );
			if ( staticDep == nullptr )
			{
				staticDep = ng.CreateFileNode( inputFiles->GetValue() );
			}
			else if ( staticDep->IsAFile() == false )
			{
				Error::Error_1103_NotAFile( funcStartIter, this, "CompilerInputFiles", staticDep->GetName(), staticDep->GetType() );
				return false;
			}
			staticDeps.Append( staticDep );
		}
		else if ( inputFiles->IsArray() )
		{
			const Array< AString > & files = inputFiles->GetArray();
			for ( Array< AString >::Iter it = files.Begin();
				  it != files.End();
				  it++ )
			{
				Node * staticDep = ng.FindNode( *it );
				if ( staticDep == nullptr )
				{
					staticDep = ng.CreateFileNode( *it );
				}
				else if ( staticDep->IsAFile() == false )
				{
					Error::Error_1103_NotAFile( funcStartIter, this, "CompilerInputFiles", staticDep->GetName(), staticDep->GetType() );
					return false;
				}
				staticDeps.Append( staticDep );
			}
		}
	}

	// do we want to build files via a unity blob?
	const BFFVariable * inputUnity = nullptr;
	if ( !GetString( funcStartIter, inputUnity, ".CompilerInputUnity" ) )
	{
		return false;
	}
	if ( inputUnity )
	{
		Node * n = ng.FindNode( inputUnity->GetValue() );
		if ( n == nullptr )
		{
			Error::Error_1104_TargetNotDefined( funcStartIter, this, "CompilerInputUnity", inputUnity->GetValue() );
			return false;
		}
		if ( n->GetType() != Node::UNITY_NODE )
		{
			Error::Error_1102_UnexpectedType( funcStartIter, this, "CompilerInputUnity", inputUnity->GetValue(), n->GetType(), Node::UNITY_NODE );
			return false;
		}
		staticDeps.Append( n );

/*		// check that precompiled header usage of unity node matches the libary
		const AString & pchLib = precompiledHeaderNode ? precompiledHeaderNode->GetName() : AString::GetEmpty();
		const AString & pchUnity = n->CastTo< UnityNode >()->GetPrecompiledHeader();
		if ( pchLib != pchUnity )
		{
			Error::Error_1302_UnityPCHMismatch( funcStartIter, this, pchLib, pchUnity );
			return false;
		}*/
	}

	if ( staticDeps.IsEmpty() )
	{
		Error::Error_1006_NothingToBuild( funcStartIter, this );
		return false;
	}

	// Create library node which depends on the single file or list
	if ( ng.FindNode( outputLib->GetValue() ) )
	{
		Error::Error_1100_AlreadyDefined( funcStartIter, this, outputLib->GetValue() );
		return false;
	}
	Node * libNode = ng.CreateLibraryNode( outputLib->GetValue(),
						  staticDeps,
						  compilerNode,
						  compilerOptions->GetValue(),
						  compilerOutputPath->GetValue(),
						  librarian->GetValue(),
						  librarianOptions->GetValue(),
						  precompiledHeaderNode );

	return ProcessAlias( funcStartIter, libNode );
}

//------------------------------------------------------------------------------
